!=========================================================================
! Copyright (C) GemTalk Systems 1986-2016.  All Rights Reserved.
!
! $Id: pathevaluator.gs,v 1.7 2008-01-09 19:02:35 stever Exp $
!
! Superclass Hierarchy:
!   PathEvaluator, Array, SequenceableCollection, Collection, Object.
!
! class created in idxclasses.topaz
!=========================================================================

removeallmethods PathEvaluator
removeallclassmethods PathEvaluator
protectmethods

category: 'For Documentation Installation only'
classmethod: PathEvaluator
installDocumentation

< unprotected >

| doc |
doc := GsClassDocumentation _newForPrivateGsClass: self.

doc documentClassWith: (GsDocText new).

self description: doc.
%

! ------------------- Class methods for PathEvaluator
category: 'Instance Creation'
classmethod: PathEvaluator
new

"Create a new initialized instance of the receiver."
<protected>

^ super new initialize
%

! ------------------- Instance methods for PathEvaluator
category: 'Initialization'
method: PathEvaluator
initialize

"Initialize a new instance of the receiver."
<protected>

ivOffsets := Array new.
constraintClasses := Array new
%

category: 'Query Select'
method: PathEvaluator
findAllValuesEqualTo: aValue

"Returns a set of all values that satisfy the query."
<protected>

^ self findAllValuesWithComparison: #= to: aValue 
%

category: 'Query Select'
method: PathEvaluator
findAllValuesGreaterThanKey: aValue andEquals: aBoolean

"Returns a set of all values that satisfy the query."
<protected>

^ self 
    findAllValuesWithComparison:
        (aBoolean
            ifTrue: [ #_idxCompareGreaterThanOrEqual: ]
            ifFalse: [ #_idxCompareGreaterThan: ])
    to: aValue 
%

category: 'Query Select'
method: PathEvaluator
findAllValuesIdenticalTo: aValue

"Returns a set of all values that satisfy the query."
<protected>

" optimize query for objects in the NSC itself "
(self at: 1) ==  #'' 
    ifTrue: [ ^ self _allIdenticalTo: aValue in: nsc ].

^ self findAllValuesWithComparison:  #== to: aValue 
%

category: 'Query Select'
method: PathEvaluator
findAllValuesLessThanKey: aValue andEquals: aBoolean

"Returns a set of all values that satisfy the query."
<protected>

^ self 
    findAllValuesWithComparison:
        (aBoolean
            ifTrue: [ #_idxCompareLessThanOrEqual: ]
            ifFalse: [ #_idxCompareLessThan: ])
    to: aValue 
%

category: 'Query Select'
method: PathEvaluator
findAllValuesNotEqualTo: aValue

"Returns a set of all values that satisfy the query."
<protected>

^ self findAllValuesWithComparison:  #~= to: aValue 
%

category: 'Query Select'
method: PathEvaluator
findAllValuesNotIdenticalTo: aValue

"Returns a set of all values that satisfy the query."
<protected>
|result|

" optimize query for objects in the NSC itself "
(self at: 1) =  #'' ifTrue: [
  result := nsc _asIdentityBag copy  .
  result removeAll: (self _allIdenticalTo: aValue in: nsc) .
  result _class = nsc speciesForSelect
    ifFalse:
      [|tmpResult |
      tmpResult := nsc speciesForSelect new.
      tmpResult addAll: result.
      result := tmpResult].
  ^ result
  ].

^ self
    findAllValuesWithComparison:  #~~
    to: aValue
%

category: 'Query Detect'
method: PathEvaluator
findFirstValueEqualTo: aValue

"Returns the first value that satisfies the query."
<protected>

| obj bag |
bag := nsc _asIdentityBag.
self isIndexOnRootNsc
    ifTrue: [
        1 to: bag size do: [ :i |
            ( (obj := bag _at: i) = aValue)
                ifTrue: [ ^ obj ]
        ]
    ]
    ifFalse: [
        1 to: bag size do: [ :i |
            nil == (obj := bag _at: i)
                ifFalse: [
                    (self _traverseObject: obj) = aValue
                        ifTrue: [ ^ obj ]
                ]
        ]
    ].

^ #_incompletePathTraversal
%

category: 'Query Detect'
method: PathEvaluator
findFirstValueGreaterThanKey: aValue
andEquals: aBoolean

"Returns the first value that satisfies the query."
<protected>

^ self
    findFirstValueWithComparison:
        (aBoolean
            ifTrue: [ #_idxCompareGreaterThanOrEqual: ]
            ifFalse: [ #_idxCompareGreaterThan: ])
    to: aValue 
%

category: 'Query Detect'
method: PathEvaluator
findFirstValueIdenticalTo: aValue

"Returns the first value that satisfies the query."
<protected>

| bag incomplete obj |
" optimize query for objects in the NSC itself "
self isIndexOnRootNsc
    ifTrue: [
        (nsc includesIdentical: aValue)
            ifTrue: [ ^ aValue ]
            ifFalse: [ ^ #_incompletePathTraversal ]
    ]
    ifFalse: [
        bag := nsc _asIdentityBag.
        incomplete := #_incompletePathTraversal.

        1 to: bag size do: [ :i |
            nil == (obj := bag _at: i)
                ifFalse: [
                    aValue == (self _traverseObject: obj)
                        ifTrue: [ ^ obj ]
                ]
        ]
    ].

^ #_incompletePathTraversal
%

category: 'Query Detect'
method: PathEvaluator
findFirstValueLessThanKey: aValue
andEquals: aBoolean

"Returns the first value that satisfies the query."
<protected>

^ self
    findFirstValueWithComparison:
        (aBoolean
            ifTrue: [ #_idxCompareLessThanOrEqual: ]
            ifFalse: [ #_idxCompareLessThan: ])
    to: aValue 
%

category: 'Query Detect'
method: PathEvaluator
findFirstValueNotEqualTo: aValue

"Returns the first value that satisfies the query."
<protected>

^ self findFirstValueWithComparison: #~= to: aValue
%

category: 'Query Detect'
method: PathEvaluator
findFirstValueNotIdenticalTo: aValue

"Returns the first value that satisfies the query."
<protected>

^ self findFirstValueWithComparison: #~~ to: aValue
%

category: 'Testing'
method: PathEvaluator
_hasSetValuedTerm

"Returns true if the path has a term that indicates a set-valued instance
 variable."
<protected>

1 to: self size do: [ :i |
    (self at: i) == #*  ifTrue: [ ^ true ]
].
^ false
%

category: 'Testing'
method: PathEvaluator
hasSetValuedTerm

"Returns true if the path has a term that indicates a set-valued instance
 variable."
<unprotected>

^ false
%

category: 'Testing'
method: PathEvaluator
isIdentityIndex

"Returns false."
<unprotected>

^ false
%

category: 'Testing'
method: PathEvaluator
isPathEvaluator

"Returns true."
<unprotected>

^ true
%

category: 'Testing'
method: PathEvaluator
isRangeEqualityIndex

"Returns true.  A path evaluator can perform the searches of an equality
 index."
<unprotected>

^ true
%

category: 'Accessing'
method: PathEvaluator
ivOffsets

"Returns the value of the instance variable 'ivOffsets'."
<protected>

^ivOffsets
%

category: 'Updating'
method: PathEvaluator
ivOffsets: newValue

"Modify the value of the instance variable 'ivOffsets'."
<protected>

ivOffsets := newValue
%

category: 'Accessing'
method: PathEvaluator
constraintClasses

"Returns the value of the instance variable 'constraintClasses'."
<protected>

^constraintClasses
%

category: 'Updating'
method: PathEvaluator
constraintClasses: newValue

"Modify the value of the instance variable 'constraintClasses'."
<protected>

constraintClasses := newValue
%

category: 'Accessing'
method: PathEvaluator
nsc

"Returns the value of the instance variable 'nsc'."
<protected>

^nsc
%

category: 'Accessing'
method: PathEvaluator
nscRoot

"Returns the value of the instance variable 'nsc'."
<protected>

^nsc
%

category: 'Updating'
method: PathEvaluator
nsc: newValue

"Modify the value of the instance variable 'nsc'."
<protected>

nsc := newValue
%

category: 'Traversing'
method: PathEvaluator
traverseObject: anObject

"Traverse the sub-objects of the given object using the path names of the
 receiver.  This method assumes that the index is not over a set-valued
 instance variable.  For that kind of index, use the traverse:startingAt:into:
 method.  If a nil value is reached before the end of the path, returns the
 incompletePathTraversalObject."
<protected>

self isIndexOnNscElements
    ifTrue: [ ^ anObject ].

nil == anObject
    ifTrue: [ ^ #_incompletePathTraversal ].

^ self _traverseObject: anObject
%

category: 'Testing'
method: PathEvaluator
isIndexOnNscElements

"Returns whether the receiver is a path directly on the elements of an NSC
 (either the root NSC or a set-valued instance variable along the path)."
<protected>

^ (self at: self size) ==  #'' 
%

category: 'Testing'
method: PathEvaluator
isIndexOnRootNsc

"Returns whether the receiver is a path directly on the elements of the root
 NSC."
<protected>

^ (self at: self size) ==  #'' 
%

category: 'Testing'
method: PathEvaluator
_isPathConstrained

"Returns true if the path represented by the receiver is fully constrained."
<protected>

^ (ivOffsets at: ivOffsets size) ~~ nil
%

category: 'Testing'
method: PathEvaluator
isPathConstrained

"Returns true if the path represented by the receiver is fully constrained."
<unprotected>

^ false
%

category: 'Traversing'
method: PathEvaluator
_traverseObject: anObject

"Traverse the sub-objects of the given object using the path names of the
 receiver.  This method assumes that the index is not over a set-valued
 instance variable.  For that kind of index, use the traverse:startingAt:into:
 method.  If a nil value is reached before the end of the path, returns the
 incompletePathTraversalObject."
<protected>

| nextObj sz |

nextObj := anObject.
sz := self size.
1 to: sz do: [ :i |
  nextObj := self _nextObjectFor: nextObj at: i.

  ( nil == nextObj _and: [ i ~= sz ] )
    ifTrue: [ ^ #_incompletePathTraversal ].
].
^ nextObj
%

category: 'Query Support'
method: PathEvaluator
_calculateIvOffsets

"Get the constraint class for each path name along the path in the receiver.
 For each constraint class, determine the instance variable offset for the
 corresponding path name, and place the offset in the receiver's Array of
 instance variable offsets.  Returns the receiver."
<protected>

^ self _calculateIvOffsetsOnKind: nsc class elementConstraint
%

category: 'Query Support'
method: PathEvaluator
_calculateIvOffsetsOnKind: aClass

"Get the constraint class for each path name along the path in the receiver,
 starting with aClass.  For each constraint class, determine the instance
 variable offset for the corresponding path name, and place the offset in
 the receiver's Array of instance variable offsets.  Returns the receiver."
<protected>

| class |
self _getConstraintsStartingOnKind: aClass.

1 to: constraintClasses size do: [ :i |
    (class := constraintClasses at: i) ~~ nil
        ifTrue: [ ivOffsets at: i put: (class _ivOffsetOf: (self at: i)) ]
        ifFalse: [ ivOffsets at: i put: nil ]
]
%

category: 'Query Select'
method: PathEvaluator
findAllValuesWithComparison: selector to: aValue

"Returns a set of all values that satisfy the query.  Must iterate through all
 elements in the NSC, traversing down the path for each object."
<protected>

" if the object at the end of the path is a basic class, then check if
we should consider nil's "
^ ( ( QueryExecuter nilOrderDefinedFor: self lastElementClass) _and:
  [ self _doNotIncludeNilsForOperator: selector ] )
    ifTrue: [ self _findAllNonNilValuesWithComparison: selector to: aValue ]
    ifFalse: [ self _findAllValuesWithComparison: selector to: aValue ]
%

category: 'Query Select'
method: PathEvaluator
_findAllValuesWithComparison: selector to: aValue

""
<protected>

| obj endObj tmpHolder bag incomplete |
tmpHolder := NscBuilder for: nsc speciesForSelect new max: self sizeForNscBuilder.

self isIndexOnRootNsc
    ifTrue: [
        bag := nsc _asIdentityBag.
        1 to: bag size do: [ :i |
            ( (obj := bag _at: i) perform: selector with: aValue)
                ifTrue: [ tmpHolder add: obj ]
        ]
    ]
    ifFalse: [
        incomplete := #_incompletePathTraversal.
        bag := nsc _asIdentityBag.
        1 to: bag size do: [ :i |
            nil == (obj := bag _at: i)
                ifFalse: [
                    endObj := self _traverseObject: obj.
                    ( endObj ~~ incomplete _and:
                    [ endObj perform: selector with: aValue ] )
                        ifTrue: [ tmpHolder add: obj ]
                ]
        ]
    ].
^ tmpHolder completeBag
%

category: 'Query Detect'
method: PathEvaluator
findFirstValueWithComparison: selector to: aValue

"Returns the first value that satisfies the query.  Must iterate through the
 elements in the NSC, traversing down the path for each object."
<protected>

( ( QueryExecuter nilOrderDefinedFor: self lastElementClass) _and:
[ self _doNotIncludeNilsForOperator: selector ] )
    ifTrue: [
        ^ self
            _findFirstNonNilValueWithComparison: selector
            to: aValue
    ]
    ifFalse: [
        ^ self
            _findFirstValueWithComparison: selector
            to: aValue
    ]
%

category: 'Query Detect'
method: PathEvaluator
_findFirstValueWithComparison: selector to: aValue

""
<protected>

| obj endObj bag incomplete |
self isIndexOnRootNsc
    ifTrue: [
        bag := nsc _asIdentityBag.
        1 to: bag size do: [ :i |
            ( (obj := bag _at: i) perform: selector with: aValue)
                ifTrue: [ ^ obj ]
        ]
    ]
    ifFalse: [
        incomplete := #_incompletePathTraversal.
        bag := nsc _asIdentityBag.
        1 to: bag size do: [ :i |
            nil == (obj := bag _at: i)
                ifFalse: [
                    endObj := self _traverseObject: obj.
                    ( endObj ~~ incomplete _and:
                    [ endObj perform: selector with: aValue ] )
                        ifTrue: [ ^ obj ]
                ]
        ]
    ].
^ #_incompletePathTraversal
%

category: 'Query Select'
method: PathEvaluator
_findAllNonNilValuesWithComparison: selector to: aValue

""
<protected>

| obj endObj tmpHolder bag incomplete |
tmpHolder := NscBuilder for: nsc speciesForSelect new max: self sizeForNscBuilder.

self isIndexOnRootNsc
    ifTrue: [
        bag := nsc _asIdentityBag.
        1 to: bag size do: [ :i |
            ( nil ~~ (obj := bag _at: i) _and:
            [ obj perform: selector with: aValue ] )
                ifTrue: [ tmpHolder add: obj ]
        ]
    ]
    ifFalse: [
        incomplete := #_incompletePathTraversal.
        bag := nsc _asIdentityBag.
        1 to: bag size do: [ :i |
            nil == (obj := bag _at: i)
                ifFalse: [
                    endObj := self _traverseObject: obj.
                    ( nil ~~ endObj _and:
                    [ (endObj ~~ incomplete) _and:
                    [ endObj perform: selector with: aValue ] ] )
                        ifTrue: [ tmpHolder add: obj ]
                ]
          ]
    ].
^ tmpHolder completeBag
%

category: 'Query Detect'
method: PathEvaluator
_findFirstNonNilValueWithComparison: selector to: aValue

""
<protected>

| obj endObj bag incomplete |
self isIndexOnRootNsc
    ifTrue: [
        bag := nsc _asIdentityBag.
        1 to: bag size do: [ :i |
            ( nil ~~ (obj := bag _at: i) _and:
            [ obj perform: selector with: aValue ] )
                ifTrue: [ ^ obj ]
        ]
    ]
    ifFalse: [
        incomplete := #_incompletePathTraversal.
        bag := nsc _asIdentityBag.
        1 to: bag size do: [ :i |
            nil == (obj := bag _at: i)
                ifFalse: [
                    endObj := self _traverseObject: obj.
                    ( nil ~~ endObj _and:
                    [ (endObj ~~ incomplete) _and:
                    [ endObj perform: selector with: aValue ] ] )
                        ifTrue: [ ^ obj ]

                ]
          ]
    ].
^ #_incompletePathTraversal
%

category: 'Testing'
method: PathEvaluator
checkAuthorization

"Determine whether the current user has read authorization on all segments for
 each object along the path.  For PathEvaluators, this operation does nothing
 because authorization is checked by actually touching the affected objects."
<unprotected>
%

category: 'Sorting'
method: PathEvaluator
_sortOn: indexObjs directions: theBooleans

"Returns an Array of objects from the receiver's NSC that is sorted according
 to the paths traversed by indexObjs (with the direction of the sort specified
 in theBooleans)."

<protected>

| sorter obj aColl firstIncompleteArray sortNodes
  hasSecondarySort incompletesArrays incomplete |

sorter := PathSorter on: indexObjs directions: theBooleans.
sortNodes := sorter sortNodes.
hasSecondarySort := indexObjs size > 1.
incomplete := #_incompletePathTraversal.

incompletesArrays := sorter incompletes.
firstIncompleteArray := incompletesArrays at: 1.

aColl := nsc _asCollectionForSorting.

" optimize when no traversal is needed "
self isIndexOnRootNsc
  ifTrue: [
    1 to: aColl size do: [ :i |
      obj := aColl _at: i.
      obj == nil
        ifTrue: [ firstIncompleteArray add: nil ]
        ifFalse: [
          sorter _addObject: obj
            inNodes: sortNodes
            hasSecondarySort: hasSecondarySort
        ]
    ].
  ]
  ifFalse: [
    1 to: aColl size do: [ :i |
      obj := aColl _at: i.
      obj == nil
        ifTrue: [ firstIncompleteArray add: nil ]
        ifFalse: [
          sorter _addObject: obj
            inNodes: sortNodes
            hasSecondarySort: hasSecondarySort
            incompletes: incompletesArrays
            incomplete: incomplete
        ]
    ]
  ].

^ sorter sortInto: (Array new: nsc size) startingAt: 1
%

category: 'Accessing'
method: PathEvaluator
lastElementClass

"Returns the class of the last object along the path (can only be determined if
 the path is fully constrained).  Returns nil if the path is not fully
 constrained."
<protected>

| lastClass |
lastClass := constraintClasses at: constraintClasses size.
lastClass == nil
    ifTrue: [ ^ nil ].

self isIndexOnRootNsc
    ifTrue: [ ^ lastClass ].

^ lastClass _constraintOn: (self at: self size)
%

category: 'Accessing'
method: PathEvaluator
pathComponentsString

"Returns a string of the path components."
<protected>

| str sz |
sz := self size.
str := String new.
1 to: sz do: [ :i |
    str addAll: (self at: i).
    i == sz
        ifFalse: [ str add: $.  ]
].
^ str

%

category: 'Query Support'
method: PathEvaluator
_allIdenticalTo: anObject in: aBag

"Returns a new NSC containing the given object with the number of occurrences
 as in the given NSC."
<protected>

| result tmp |
result := self _primAllIdenticalTo: anObject in: aBag _asIdentityBag.
aBag _isIdentityBag
  ifFalse: [
    tmp := aBag class new: result size.
    tmp addAll: result.
    result := tmp
  ].
^ result
%

category: 'Query Support'
method: PathEvaluator
_primAllIdenticalTo: anObject in: aBag

"Returns a new NSC containing the given object with the number of occurrences as
 in the given NSC."

<protected primitive: 340>
self _primitiveFailed: #allIdenticalTo:in: .
self _uncontinuableError
%

category: 'Query Support'
method: PathEvaluator
hasSamePathAs: aPathEvaluator

"Returns whether the given path evaluator is on the same path as the receiver."
<protected>

" make sure it is a path evaluator "
aPathEvaluator isPathEvaluator
    ifFalse: [ ^ false ].

" make sure they have the same number of terms "
self size == aPathEvaluator size
    ifFalse: [ ^ false ].

" check each term "
1 to: self size do: [ :i |
    (self at: i) == (aPathEvaluator at: i)
        ifFalse: [ ^ false ]
].
^ true
%

category: 'Query Support'
method: PathEvaluator
_doNotIncludeNilsForOperator: selector

"Returns whether nil values at the end of the path should be considered for the
 given operator (a message selector)."
<protected>

 #_idxCompareLessThan: ==  selector ifTrue:[ ^ true].
 #_idxCompareLessThanOrEqual: == selector ifTrue:[ ^ true ].
 #_idxCompareGreaterThan: == selector ifTrue:[ ^ true ].
 #_idxCompareGreaterThanOrEqual: == selector ifTrue:[ ^ true ].
 ^ false 
%

category: 'Query Support'
method: PathEvaluator
_getConstraintsStartingOnKind: aClass

"Calculate the constraint Array based upon the path beginning with
 instances of aClass."
<protected>

| tmpClass |
" if path is empty string, then use given class "
(self at: 1) ==  #'' 
    ifTrue: [
        constraintClasses at: 1 put: aClass.
        ^ self
    ].

tmpClass := aClass.
1 to: self size do: [ :i |
    constraintClasses at: i put: tmpClass.
    tmpClass == nil
        ifFalse: [
            tmpClass isNsc
                ifTrue: [ tmpClass := tmpClass elementConstraint ]
                ifFalse: [ tmpClass := (tmpClass _constraintOn: (self at: i)) ]
        ].
]
%

category: 'Traversing'
method: PathEvaluator
_traverseObject: anObject
incompletesInto: incompleteArray
incomplete: incomplete

"Traverse the sub-objects of the given object using the path names of the
 receiver.  This method assumes that the index is not over a set-valued
 instance variable.  For that kind of index, use the traverse:startingAt:into:
 method.  If a nil value is reached before the end of the path, place the object
 into the appropriate Array in the incompleteArray of Arrays."
<protected>

| nextObj sz |
nextObj := anObject.
sz := self size.
1 to: sz do: [ :i |
  nextObj := self _nextObjectFor: nextObj at: i.

  ( nil == nextObj _and: [ i ~= sz ] )
    ifTrue: [
      incompleteArray == nil
        ifTrue: [ ^ nil ].

      (incompleteArray at: i + 1) add: anObject.
        ^ incomplete
    ].
].
^ nextObj
%

category: 'Traversing'
method: PathEvaluator
traverseObject: anObject
incompletesInto: incompleteArray
incomplete: incomplete

"Traverse the sub-objects of the given object using the path names of the
 receiver.  This method assumes that the index is not over a set-valued
 instance variable.  For that kind of index, use the traverse:startingAt:into:
 method.  If a nil value is reached before the end of the path, place the object
 into the appropriate Array in the incompleteArray of Arrays."
<protected>

self isIndexOnNscElements
    ifTrue: [ ^ anObject ].

nil == anObject
    ifTrue: [ ^ incomplete ].

^ self _traverseObject: anObject
  incompletesInto: incompleteArray
  incomplete: incomplete
%

category: 'Query Select'
method: PathEvaluator
findAllValuesGreaterThan: val1
andEquals: bool1
andLessThan: val2
andEquals: bool2

""
<protected>

| obj endObj tmpHolder bag incomplete |
" if query on NSC elements, execute optimized method (no traversal) "
self isIndexOnRootNsc
  ifTrue: [
    ^ self _findAllValuesGreaterThan: val1
      andEquals: bool1
      andLessThan: val2
      andEquals: bool2
  ].

tmpHolder := NscBuilder for: nsc speciesForSelect new max: self sizeForNscBuilder.

incomplete := #_incompletePathTraversal.
bag := nsc _asIdentityBag.

bool1
  ifTrue: [
    bool2
      ifTrue: [
        1 to: bag size do: [ :i |
          nil == (obj := bag _at: i)
            ifFalse: [
              endObj := self _traverseObject: obj.

              ( endObj ~~ incomplete _and:
              [ ( endObj _idxCompareGreaterThanOrEqual: val1 ) _and:
              [ endObj _idxCompareLessThanOrEqual: val2 ] ] )
                ifTrue: [ tmpHolder add: obj ]
            ]
        ]
      ]
      ifFalse: [ " bool2 is false "
        1 to: bag size do: [ :i |
          nil == (obj := bag _at: i)
            ifFalse: [
              endObj := self _traverseObject: obj.

              ( endObj ~~ incomplete _and:
              [ ( endObj _idxCompareGreaterThanOrEqual: val1 ) _and:
              [ endObj _idxCompareLessThan: val2 ] ] )
                ifTrue: [ tmpHolder add: obj ]
            ]
        ]
      ]
  ]
  ifFalse: [ " bool1 is false "
    bool2
      ifTrue: [
        1 to: bag size do: [ :i |
          nil == (obj := bag _at: i)
            ifFalse: [
              endObj := self _traverseObject: obj.

              ( endObj ~~ incomplete _and:
              [ ( endObj _idxCompareGreaterThan: val1 ) _and:
              [ endObj _idxCompareLessThanOrEqual: val2 ] ] )
                ifTrue: [ tmpHolder add: obj ]
            ]
        ]
      ]
      ifFalse: [ " bool2 is false "
        1 to: bag size do: [ :i |
          nil == (obj := bag _at: i)
            ifFalse: [
              endObj := self _traverseObject: obj.

              ( endObj ~~ incomplete _and:
              [ ( endObj _idxCompareGreaterThan: val1 ) _and:
              [ endObj _idxCompareLessThan: val2 ] ] )
                ifTrue: [ tmpHolder add: obj ]
            ]
        ]
      ]
  ].

^ tmpHolder completeBag
%

category: 'Query Detect'
method: PathEvaluator
findFirstValueGreaterThan: val1
andEquals: bool1
andLessThan: val2
andEquals: bool2

""
<protected>

| obj endObj bag incomplete |
incomplete := #_incompletePathTraversal.
bag := nsc _asIdentityBag.

bool1
  ifTrue: [
    bool2
      ifTrue: [
        1 to: bag size do: [ :i |
          nil == (obj := bag _at: i)
            ifFalse: [
              endObj := self _traverseObject: obj.

              ( endObj ~~ incomplete _and:
              [ ( endObj _idxCompareGreaterThanOrEqual: val1 ) _and:
              [ endObj _idxCompareLessThanOrEqual: val2 ] ] )
                ifTrue: [ ^ obj ]
            ]
        ]
      ]
      ifFalse: [ " bool2 is false "
        1 to: bag size do: [ :i |
          nil == (obj := bag _at: i)
            ifFalse: [
              endObj := self _traverseObject: obj.

              ( endObj ~~ incomplete _and:
              [ ( endObj _idxCompareGreaterThanOrEqual: val1 ) _and:
              [ endObj _idxCompareLessThan: val2 ] ] )
                ifTrue: [ ^ obj ]
            ]
        ]
      ]
  ]
  ifFalse: [ " bool1 is false "
    bool2
      ifTrue: [
        1 to: bag size do: [ :i |
          nil == (obj := bag _at: i)
            ifFalse: [
              endObj := self _traverseObject: obj.

              ( endObj ~~ incomplete _and:
              [ ( endObj _idxCompareGreaterThan: val1 ) _and:
              [ endObj _idxCompareLessThanOrEqual: val2 ] ] )
                ifTrue: [ ^ obj ]
            ]
        ]
      ]
      ifFalse: [ " bool2 is false "
        1 to: bag size do: [ :i |
          nil == (obj := bag _at: i)
            ifFalse: [
              endObj := self _traverseObject: obj.

              ( endObj ~~ incomplete _and:
              [ ( endObj _idxCompareGreaterThan: val1 ) _and:
              [ endObj _idxCompareLessThan: val2 ] ] )
                ifTrue: [ ^ obj ]
            ]
        ]
      ]
  ].

^ #_incompletePathTraversal
%

category: 'Converting'
method: PathEvaluator
asPathEvaluator

""
<unprotected>

^ self
%

category: 'Converting'
method: PathEvaluator
asMostSpecificType

"Returns a new instance of a subclass, or self, depending upon the path."
<protected>

self _calculateIvOffsets.

self _hasSetValuedTerm
    ifTrue: [ ^ self changeClassTo: SetValuedPathEvaluator ].

self _isPathConstrained
    ifTrue: [ ^ self changeClassTo: ConstrainedPathEvaluator ].
%

category: 'Query Support'
method: PathEvaluator
traverse: anObject andCompare: selector with: aValue

"Traverse the given object and perform the comparison operation
 on the object at the end of the path and the given value."
<protected>

| endObj  |

^ ( (endObj := self _traverseObject: anObject) ~~ #_incompletePathTraversal _and:
[ endObj perform: selector with: aValue ] )
%

category: 'Query Support'
method: PathEvaluator
traverse: anObject andCompare: selector withTraverser: aTraverser

"Traverse the given object and perform the comparison operation on the object
 at the end of the path and the value obtained by using the traverser."
<protected>

| endObj1 endObj2  |

^ (endObj1 := self _traverseObject: anObject) ~~ #_incompletePathTraversal _and:
[ (endObj2 := aTraverser _traverseObject: anObject) ~~ #_incompletePathTraversal _and:
[ endObj1 perform: selector with: endObj2 ] ]
%

category: 'Accessing'
method: PathEvaluator
sizeForNscBuilder

"Returns the size to be used when an NscBuilder is collecting objects
 based upon the receiver's path."
<protected>

^ self hasSetValuedTerm
  ifTrue: [ NscBuilder maxSize ]
  ifFalse: [ nsc size ]
%

category: 'Traversing'
method: PathEvaluator
_nextObjectFor: anObject at: offset

"Returns the next object along the path in anObject for the named instance
 variable at the given offset in the receiver."
<protected>

| ivOffset |
anObject class == (constraintClasses at: offset)
  ifTrue: [ ivOffset := ivOffsets at: offset ]
  ifFalse: [ ivOffset := anObject class _ivOffsetOf: (self at: offset) ].

ivOffset == nil
  ifTrue: [ anObject _errorInvalidOffset: (self at: offset) ].

^ anObject instVarAt: ivOffset
%

category: 'Traversing'
method: PathEvaluator
_initTraverseCache

"Do nothing."
<unprotected>

%

category: 'Query Select'
method: PathEvaluator
_findAllValuesGreaterThan: val1
andEquals: bool1
andLessThan: val2
andEquals: bool2

"This method is optimized for the case when the query is directly on
 the elements of the NSC (that is, no traversal is needed)."
<protected>

| obj tmpHolder bag |
tmpHolder := NscBuilder for: nsc speciesForSelect new max: self sizeForNscBuilder.

bag := nsc _asIdentityBag.

bool1
  ifTrue: [
    bool2
      ifTrue: [
        1 to: bag size do: [ :i |
          nil == (obj := bag _at: i)
            ifFalse: [
              ( ( obj _idxCompareGreaterThanOrEqual: val1 ) _and:
              [ obj _idxCompareLessThanOrEqual: val2 ] )
                ifTrue: [ tmpHolder add: obj ]
            ]
        ]
      ]
      ifFalse: [ " bool2 is false "
        1 to: bag size do: [ :i |
          nil == (obj := bag _at: i)
            ifFalse: [
              ( ( obj _idxCompareGreaterThanOrEqual: val1 ) _and:
              [ obj _idxCompareLessThan: val2 ] )
                ifTrue: [ tmpHolder add: obj ]
            ]
        ]
      ]
  ]
  ifFalse: [ " bool1 is false "
    bool2
      ifTrue: [
        1 to: bag size do: [ :i |
          nil == (obj := bag _at: i)
            ifFalse: [
              ( ( obj _idxCompareGreaterThan: val1 ) _and:
              [ obj _idxCompareLessThanOrEqual: val2 ] )
                ifTrue: [ tmpHolder add: obj ]
            ]
        ]
      ]
      ifFalse: [ " bool2 is false "
        1 to: bag size do: [ :i |
          nil == (obj := bag _at: i)
            ifFalse: [
              ( ( obj _idxCompareGreaterThan: val1 ) _and:
              [ obj _idxCompareLessThan: val2 ] )
                ifTrue: [ tmpHolder add: obj ]
            ]
        ]
      ]
  ].

^ tmpHolder completeBag
%

unprotectmethods
